"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
Object.defineProperty(exports, "AnalysisError", {
  enumerable: true,
  get: function () {
    return _error.AnalysisError;
  }
});
exports.analyzeBody = analyzeBody;
Object.defineProperty(exports, "analyzeHead", {
  enumerable: true,
  get: function () {
    return _head.analyzeHead;
  }
});
exports.needsProps = needsProps;
var _error = require("./analysis/error.js");
var _this_fields = require("./analysis/this_fields.js");
var _state = require("./analysis/state.js");
var _utils = require("./utils.js");
var _prop = require("./analysis/prop.js");
var _local = require("./analysis/local.js");
var _user_defined = require("./analysis/user_defined.js");
var _head = require("./analysis/head.js");
const SPECIAL_STATIC_NAMES = new Set(["childContextTypes", "contextTypes", "contextType", "defaultProps", "getDerivedStateFromError", "getDerivedStateFromProps"]);
function analyzeBody(path, head) {
  const locals = new _local.LocalManager(path);
  const {
    thisFields: sites,
    staticFields
  } = (0, _this_fields.analyzeThisFields)(path);
  const propsObjSites = (0, _utils.getAndDelete)(sites, "props") ?? [];
  const defaultPropsObjSites = (0, _utils.getAndDelete)(staticFields, "defaultProps") ?? [];
  const stateObjSites = (0, _utils.getAndDelete)(sites, "state") ?? [];
  const setStateSites = (0, _utils.getAndDelete)(sites, "setState") ?? [];
  const states = (0, _state.analyzeState)(stateObjSites, setStateSites, locals, head);
  const renderSites = (0, _utils.getAndDelete)(sites, "render") ?? [];
  analyzeOuterCapturings(path, locals);
  let renderPath = undefined;
  {
    if (renderSites.some(site => site.type === "expr")) {
      throw new _error.AnalysisError(`do not use this.render`);
    }
    const init = renderSites.find(site => site.init);
    if (init) {
      if (init.path.isClassMethod()) {
        renderPath = init.path;
      }
    }
  }
  const userDefined = (0, _user_defined.analyzeUserDefined)(sites);
  for (const [name] of staticFields) {
    if (!SPECIAL_STATIC_NAMES.has(name)) {
      throw new _error.AnalysisError(`Cannot transform static ${name}`);
    } else {
      throw new _error.AnalysisError(`Cannot transform static ${name}`);
    }
  }
  if (!renderPath) {
    throw new _error.AnalysisError(`Missing render method`);
  }
  const props = (0, _prop.analyzeProps)(propsObjSites, defaultPropsObjSites, locals, head);
  for (const [name, propAnalysis] of props.props) {
    if (propAnalysis.needsAlias) {
      propAnalysis.newAliasName = locals.newLocal(name, propAnalysis.sites.map(site => site.path));
    }
  }
  const render = analyzeRender(renderPath, locals);
  for (const [name, stateAnalysis] of states.entries()) {
    const bindingPaths = stateAnalysis.sites.map(site => site.path);
    stateAnalysis.localName = locals.newLocal(name, bindingPaths);
    stateAnalysis.localSetterName = locals.newLocal(`set${name.replace(/^[a-z]/, s => s.toUpperCase())}`, bindingPaths);
  }
  for (const [name, field] of userDefined.fields) {
    field.localName = locals.newLocal(name, field.sites.map(site => site.path));
  }
  return {
    locals,
    render,
    state: states,
    props,
    userDefined
  };
}
function analyzeRender(path, locals) {
  const renames = [];
  for (const [name, binding] of Object.entries(path.scope.bindings)) {
    if (locals.allRemovePaths.has(binding.path)) {
      // Already handled as an alias
      continue;
    }
    const newName = locals.newLocal(name, []);
    renames.push({
      scope: binding.scope,
      oldName: name,
      newName
    });
  }
  return {
    path,
    renames
  };
}
function analyzeOuterCapturings(classPath, locals) {
  const capturings = new Set();
  function visitIdent(path) {
    path.getOuterBindingIdentifiers;
    const binding = path.scope.getBinding(path.node.name);
    if (!binding || binding.path.isAncestor(classPath)) {
      capturings.add(path.node.name);
      locals.markCaptured(path.node.name);
    }
  }
  classPath.get("body").traverse({
    Identifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    },
    JSXIdentifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    }
  });
  return capturings;
}
function needsProps(body) {
  return body.props.sites.length > 0;
}