import { AnalysisError } from "./analysis/error.js";
import { analyzeThisFields } from "./analysis/this_fields.js";
import { analyzeState } from "./analysis/state.js";
import { getAndDelete } from "./utils.js";
import { analyzeProps } from "./analysis/prop.js";
import { LocalManager } from "./analysis/local.js";
import { analyzeUserDefined } from "./analysis/user_defined.js";
export { AnalysisError } from "./analysis/error.js";
export { analyzeHead } from "./analysis/head.js";
const SPECIAL_STATIC_NAMES = new Set(["childContextTypes", "contextTypes", "contextType", "defaultProps", "getDerivedStateFromError", "getDerivedStateFromProps"]);
export function analyzeBody(path, head) {
  const locals = new LocalManager(path);
  const {
    thisFields: sites,
    staticFields
  } = analyzeThisFields(path);
  const propsObjSites = getAndDelete(sites, "props") ?? [];
  const defaultPropsObjSites = getAndDelete(staticFields, "defaultProps") ?? [];
  const stateObjSites = getAndDelete(sites, "state") ?? [];
  const setStateSites = getAndDelete(sites, "setState") ?? [];
  const states = analyzeState(stateObjSites, setStateSites, locals, head);
  const renderSites = getAndDelete(sites, "render") ?? [];
  analyzeOuterCapturings(path, locals);
  let renderPath = undefined;
  {
    if (renderSites.some(site => site.type === "expr")) {
      throw new AnalysisError(`do not use this.render`);
    }
    const init = renderSites.find(site => site.init);
    if (init) {
      if (init.path.isClassMethod()) {
        renderPath = init.path;
      }
    }
  }
  const userDefined = analyzeUserDefined(sites);
  for (const [name] of staticFields) {
    if (!SPECIAL_STATIC_NAMES.has(name)) {
      throw new AnalysisError(`Cannot transform static ${name}`);
    } else {
      throw new AnalysisError(`Cannot transform static ${name}`);
    }
  }
  if (!renderPath) {
    throw new AnalysisError(`Missing render method`);
  }
  const props = analyzeProps(propsObjSites, defaultPropsObjSites, locals, head);
  for (const [name, propAnalysis] of props.props) {
    if (propAnalysis.needsAlias) {
      propAnalysis.newAliasName = locals.newLocal(name, propAnalysis.sites.map(site => site.path));
    }
  }
  const render = analyzeRender(renderPath, locals);
  for (const [name, stateAnalysis] of states.entries()) {
    const bindingPaths = stateAnalysis.sites.map(site => site.path);
    stateAnalysis.localName = locals.newLocal(name, bindingPaths);
    stateAnalysis.localSetterName = locals.newLocal(`set${name.replace(/^[a-z]/, s => s.toUpperCase())}`, bindingPaths);
  }
  for (const [name, field] of userDefined.fields) {
    field.localName = locals.newLocal(name, field.sites.map(site => site.path));
  }
  return {
    locals,
    render,
    state: states,
    props,
    userDefined
  };
}
function analyzeRender(path, locals) {
  const renames = [];
  for (const [name, binding] of Object.entries(path.scope.bindings)) {
    if (locals.allRemovePaths.has(binding.path)) {
      // Already handled as an alias
      continue;
    }
    const newName = locals.newLocal(name, []);
    renames.push({
      scope: binding.scope,
      oldName: name,
      newName
    });
  }
  return {
    path,
    renames
  };
}
function analyzeOuterCapturings(classPath, locals) {
  const capturings = new Set();
  function visitIdent(path) {
    path.getOuterBindingIdentifiers;
    const binding = path.scope.getBinding(path.node.name);
    if (!binding || binding.path.isAncestor(classPath)) {
      capturings.add(path.node.name);
      locals.markCaptured(path.node.name);
    }
  }
  classPath.get("body").traverse({
    Identifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    },
    JSXIdentifier(path) {
      if (path.isReferencedIdentifier()) {
        visitIdent(path);
      }
    }
  });
  return capturings;
}
export function needsProps(body) {
  return body.props.sites.length > 0;
}