import { getOr, memberName } from "../utils.js";
import { AnalysisError } from "./error.js";
import { trackMember } from "./track_member.js";
/**
 * Detects assignments that expand `this.props` to variables, like:
 *
 * ```js
 * const { foo, bar } = this.props;
 * ```
 *
 * or:
 *
 * ```js
 * const foo = this.props.foo;
 * const bar = this.props.bar;
 * ```
 */
export function analyzeProps(propsObjSites, defaultPropsObjSites, locals, head) {
  const defaultProps = analyzeDefaultProps(defaultPropsObjSites);
  const newObjSites = [];
  const props = new Map();
  const getProp = name => getOr(props, name, () => ({
    sites: [],
    aliases: [],
    needsAlias: false
  }));
  for (const site of propsObjSites) {
    if (site.type !== "expr" || site.hasWrite) {
      throw new AnalysisError(`Invalid use of this.props`);
    }
    newObjSites.push({
      path: site.path
    });
    const memberAnalysis = trackMember(site.path);
    if (memberAnalysis.fullyDecomposed && memberAnalysis.memberAliases) {
      for (const [name, aliasing] of memberAnalysis.memberAliases) {
        getProp(name).aliases.push({
          scope: aliasing.scope,
          localName: aliasing.localName
        });
        locals.reserveRemoval(aliasing.idPath);
      }
    } else if (defaultProps) {
      if (memberAnalysis.memberExpr) {
        getProp(memberAnalysis.memberExpr.name).sites.push({
          path: memberAnalysis.memberExpr.path
        });
      } else {
        throw new AnalysisError(`Non-analyzable this.props in presence of defaultProps`);
      }
    }
  }
  for (const [name, propTyping] of head.propsEach) {
    getProp(name).typing = propTyping;
  }
  if (defaultProps) {
    for (const [name, defaultValue] of defaultProps) {
      getProp(name).defaultValue = defaultValue;
    }
  }
  for (const [, prop] of props) {
    prop.needsAlias = prop.aliases.length > 0 || prop.sites.length > 0;
  }
  const allAliases = Array.from(props.values()).flatMap(prop => prop.aliases);
  return {
    hasDefaults: !!defaultProps,
    sites: newObjSites,
    props,
    allAliases
  };
}
function analyzeDefaultProps(defaultPropsSites) {
  for (const site of defaultPropsSites) {
    if (!site.init) {
      throw new AnalysisError(`Invalid use of static defaultState`);
    }
  }
  const defaultPropsFields = new Map();
  const init = defaultPropsSites.find(site => site.init);
  if (!init) {
    return;
  }
  const init_ = init.init;
  if (init_.type !== "init_value") {
    throw new AnalysisError("Non-analyzable defaultProps initializer");
  }
  const initPath = init_.valuePath;
  if (!initPath.isObjectExpression()) {
    throw new AnalysisError("Non-analyzable defaultProps initializer");
  }
  for (const fieldPath of initPath.get("properties")) {
    if (!fieldPath.isObjectProperty()) {
      throw new AnalysisError("Non-analyzable defaultProps initializer");
    }
    const stateName = memberName(fieldPath.node);
    if (stateName == null) {
      throw new AnalysisError("Non-analyzable defaultProps initializer");
    }
    const fieldInitPath = fieldPath.get("value");
    if (!fieldInitPath.isExpression()) {
      throw new AnalysisError("Non-analyzable defaultProps initializer");
    }
    defaultPropsFields.set(stateName, fieldInitPath);
  }
  return defaultPropsFields.size > 0 ? defaultPropsFields : undefined;
}